# /Topic/TopicModel.py
# Abstract class defining Topic model
import abc
from womg.topic.topic_model import TopicModel


class TLTTopicModel(TopicModel):
    '''
    Abstract class for topic models involved in tlt diffusion model class

    Attributes
    ----------
    viralities : dict
        key: document index, value: int virality parameter

    Methods
    -------
    topic_distrib_extraction : absract
        inferes topic distribution for each document
    '''

    def __init__(self):
        super().__init__()
        self.viralities = {}

    @abc.abstractmethod
    def get_items_descript(self):
        '''
        Methods for infering topic distributions of the given documents
        '''
        pass



# /diffusion/diffusion_model.py
# Abstract class defining Diffusion models
import abc
import pickle
import pathlib


class DiffusionModel(abc.ABC):
    '''
    Abstract class for diffusion models

    Attributes
    ----------
    network_model : networkModel obj
    topic_model : topicModel obj

    Methods
    -------
    validate_config : concrete
        validates the correct tlt components
    run : concrete
        run the iteration method steps time
    iteration : abstract
        defines the one step iteration
    stop_criterior
        defines how to stop simulation
    save_model_attr
        saves all the attributes
    save_model_class
        saves all the class in pickle file

    Notes
    -----
    Hidden attributes names has to start with "_"
    '''

    def __init__(self):
        self.network_model = {}
        self.topic_model = {}


    def run(self, numb_steps):
        '''
        Simulates by running iteration method numb_steps times

        Parameters
        ----------
        numb_steps : int
            number of simulation steps, i.e. number of times to call iteration()
        '''
        print("Computing cascades.. ")
        for t in self._progress_bar(range(numb_steps)):
            #print(t)
            if t==0:
                self.iteration(step=0)
            else:
                if not self.stop_criterior():
                    self.iteration(step=t)
                else:
                    #print('\n stop_criterior: ', self.stop_criterior(), '\n')
                    #print('\n Simulation stopped at timestep ', str(t) ,
                    print('WoMG diffusion has been completed.'  )
                    break


    @abc.abstractmethod
    def iteration(self, step):
        '''
        Abstract method for defining the iteration of the diffusion process:
        how nodes will activate on items in a fixed step
        '''
        pass

    @abc.abstractmethod
    def stop_criterior(self):
        '''
        Abstract method for defining a particular case in which simulation has
        to stop
        '''
        pass



# /Topic/lda.py
# Implementation of LDA topic-model
import re
import pathlib
import gensim
import numpy as np
from nltk.corpus import stopwords
from topic.tlt_topic_model import TLTTopicModel
from utils.utility_functions import count_files, read_docs, TopicsError
from utils.distributions import random_viralities_vec


class LDA(TLTTopicModel):
    '''
    Class implementing Latent Dirichlet Allocation as topic model
    for topic distribution involved in tlt class model

    Attributes
    ----------
    numb_topics : int
        hidden
    items_keyw : dict
        dict of items in bow format

    Methods
    -------
    set_lda_mode : concrete
        sets the lda mode: reading or generating mode
    '''

    def __init__(self, numb_topics, numb_docs, docs_path, items_descr_path):
        super().__init__()
        self.numb_topics = numb_topics
        self.numb_docs = numb_docs
        self.numb_words = 20
        self._docs_path = docs_path
        self._items_descr_path = items_descr_path
        self.items_keyw = {}


    def fit(self):
        '''
        Pipeline for fitting the lda model

        1. define the lda mode : generative mode / reading mode
        2. train lda
        3. get the items descriptions (topic distribution for each item)
        4. get the items keywords (bow list for each item)
        '''
        mode = self.set_lda_mode()
        if mode == 'load':
            self.items_descript, self.numb_docs = self.load_items_descr(self._items_descr_path)
        else:
            self.make_dictionary()
            lda_model = self.train_lda()
            if mode == 'get':
                self.topics_descript = self.get_topics_descript(lda_model)
                self.get_items_descript(path=self._docs_path, model=lda_model)
                self.get_items_keyw(path=self._docs_path, model=lda_model)
            if mode == 'gen':
                self.topics_descript = self.get_topics_descript(lda_model)
                self.gen_items_descript(model=lda_model)
                self.gen_items_keyw(model=lda_model)



    def set_lda_mode(self):
        '''
        Sets how lda has to work:
        reading a document folder or generating documents


        Parameters
        ----------
        path : string
            position of the document folder

        Returns
        -------
        reading : bool
            if True: it will read docs inside the given folder path or input folder
            if False: it will use lda for generating docs

        '''
        # setting mode
        if self.numb_docs == None and self._docs_path == None:
            mode = 'load'
            if self._items_descr_path == None:
                # pre-trained topic model with 15 topics and 50 docs
                self.items_descr_path = ' '
            else:
                # given trained topic model
                self.items_descript = self._items_descr_path

        if self.numb_docs == None and self._docs_path != None and self._items_descr_path == None:
            mode = 'get'
            path = pathlib.Path(self._docs_path)
            numb_docs = count_files(path)
            if numb_docs != 0:
                self.numb_docs = numb_docs
                print('Extracting topic distribution from docs in ', path)
            else:
                print('No txt file in: ', path)
            self._training_path = pathlib.Path.cwd().parent / "data" / "docs" / "training_corpus2"


        if self.numb_docs != None and self._docs_path == None and self._items_descr_path == None:
            mode = 'gen'
            print('Setting LDA in generative mode: ', self.numb_docs, ' documents, with ', self.numb_topics, ' topics.')
            print('Training the LDA model ..')
            self._training_path = pathlib.Path.cwd().parent / "data" / "docs" / "training_corpus2"

        return mode


    def set_docs_viralities(self, virality):
        '''
        Sets the documents viralities to the given scalar/vector

        Parameters
        ----------
        viralitiy : float
            Exponent of the powerlaw distribution for documents
            viralities. P(x; a) = x^{-a}, 0 <= x <=1
        '''
        viralities = random_viralities_vec(gamma=virality, dimensions=self.numb_docs)

        if np.size(viralities) == self.numb_docs:
            for item in range(self.numb_docs):
                self.viralities[item] = viralities[item]
        if np.size(viralities) == 1:
            for item in range(self.numb_docs):
                self.viralities[item] = viralities
        #print(viralities)


    def get_items_descript(self, path, model):
        '''
        Gets the topic distribution as attribute of the superclass
        '''
        docs = read_docs(path)
        corpus = self.preprocess_texts(docs)
        gammas = {}
        item = 0
        for item in range(self.numb_docs):
            item_descript = model.get_document_topics(corpus[item], minimum_probability=0.)
            gammas[item] = np.array([i[1] for i in item_descript])
            item += 1
        if self.numb_docs == len(gammas.keys()):
            print("Items' distribution over topics is stored")
            #print(gammas)
        self.items_descript = gammas

    def gen_items_descript(self, model):
        '''
        Generates the topic distribution for each item
        and stores it in the imets_descript attribute
        '''
        alpha = model.alpha
        #alpha = [0.01120081, 0.04134526, 0.5296952,  0.00861911, 0.00862031, 0.01053169, 0.01223436, 0.1643439,  0.00871354, 0.00967268, 0.01102241, 0.01131404, 0.0118466,  0.02180933, 0.0123167]
        gammas = {}
        for item in range(self.numb_docs):
            gammas[item] = np.random.dirichlet(alpha)
        self.items_descript = gammas


    def get_items_keyw(self, path, model):
        '''
        Get the items keyword in a bow format
        '''
        return

    def gen_items_keyw(self, model):
        '''
        Generates the items keyword in a bow format
        '''
        items_descript = self.items_descript
        topics_descript = self.get_topics_descript(model, mtrx_form=True)

        for item in range(self.numb_docs):
            item_keyw = []
            for word in range(self.numb_words):
                multi_items = np.random.multinomial(1, items_descript[item], size=1)
                topic_index = np.where(multi_items==1)[1]
                multi_topics = np.random.multinomial(1, topics_descript[topic_index][0], size=1)
                word_index = np.where(multi_topics==1)[1]
                item_keyw.append(self.dictionary[word_index[0]])
            self.items_keyw[item] = self.to_bow(item_keyw)

    def get_topics_descript(self, model, mtrx_form=False):
        '''
        Getting the word distribution for each topic

        Parameters
        ----------
        model : gensim obj
            input model from gensim library e.g. lda model trained
        all : bool
            if True the complete word distribution for each topic is given
            in a (topics)x(words) matrix format
            if False a combination of most important words is given for each topic
        '''
        if mtrx_form:
            return model.get_topics()
        else:
            return model.print_topics()


    def load_items_descr(self, path):
        '''
        Returns the items_descript loaded from a file in path

        Parameters
        ----------
        path : int
            path of the items_descript file

        Returns
        -------
        tuple of items_descript loaded from path and numb_docs
        '''
        items_descr_dict = {}
        with open(self._items_descr_path, 'r') as f:
            numb_docs = 0
            for line in f:
                line = line.replace(',','').replace(']','').replace('[','')
                values = line.split()
                node = int(values[0])
                interests_vec = [float(i) for i in values[1:]]
                if self.numb_topics != len(interests_vec):
                    raise TopicsError("Please write the correct number of topics as input")
                items_descr_dict[node] = interests_vec
                numb_docs += 1
        return items_descr_dict, numb_docs


    def to_bow(self, text):
        '''
        Returns the give doc in a bow format
        '''
        bow = self.dictionary.doc2bow(text)
        bow_dict = {}
        for word_index, freq in bow:
            bow_dict[self.dictionary[word_index]] = freq
        return bow_dict

    @staticmethod
    def clean_text(docs):
        '''

        Parameters
        ----------
        docs : list of lists
            each entry is the list containing the doc in string format

        '''
        print(docs)
        # Remove new line characters
        data = [re.sub(r'\s+', ' ', str(sent[0])) for sent in docs]

        # Remove distracting single quotes
        data = [re.sub(r"\'", "", str(sent[0])) for sent in docs]

        # Remove all the special characters
        data = [re.sub(r'\W', ' ', str(sent[0])) for sent in docs]

        # remove all single characters
        data = [re.sub(r'\s+[a-zA-Z]\s+', ' ', str(sent[0])) for sent in docs]

        # Remove single characters from the start
        data = [re.sub(r'\^[a-zA-Z]\s+', ' ', str(sent[0])) for sent in docs]

        # Substituting multiple spaces with single space
        data = [re.sub(r'\s+', ' ', str(sent[0]), flags=re.I) for sent in docs]

        # Removing prefixed 'b'
        data = [re.sub(r'^b\s+', '', str(sent[0])) for sent in docs]

        # Remove article
        data = [re.sub(r'the', '', str(sent[0])) for sent in docs]

        # Remove to
        data = [re.sub(r'to', '', str(sent[0])) for sent in docs]

        #print(data)
        return data

    def sent_to_words(self, docs):
        '''
        Returns list of docs divided in words

        Parameters
        ----------
        docs : list
            list of lists: each entry is the list containing the document as a string

        Returns
        -------
        data_words : list
            list of lists: each entry is the list of words of one doc
        '''
        for sentence in docs:
            yield(gensim.utils.simple_preprocess(str(sentence), deacc=True))


    def remove_stop_words(self, texts):
        '''
        '''
        stop_words = stopwords.words('english')
        return [[word for word in gensim.utils.simple_preprocess(str(doc)) if word not in stop_words] for doc in texts]



    def make_dictionary(self):
        '''
        Sets the dictionary attribute using a training corpus in input

        Parameters
        ----------
        data_words : list
            list of lists: each entry is the list of words of one doc

        Returns
        -------
        id2word : list
            dictionary of unique words
        '''
        train_corpus = read_docs(self._training_path)
        clean_train_corpus = self.clean_text(train_corpus)
        dictio = list(self.sent_to_words(clean_train_corpus))
        self.dictionary = gensim.corpora.Dictionary(dictio)


    def train_lda(self):
        '''
        Pre-train lda model on a saved corpus for infering the prior weights of
        the distributions given a number of topics, which correpsonds to the
        weights' dimension

        Returns gensim object of lda model

        Parameters
        ----------
        path : str
            path of the training corpus

        Returns
        -------
        gensim lda model object
        '''
        docs = read_docs(self._training_path)
        #clean_docs = self.clean_text(docs)
        docs_words = list(self.sent_to_words(docs))
        #print(docs_words)
        corpus = [self.dictionary.doc2bow(text, return_missing=True) for text in docs_words]
        #print(corpus)
        #print(self.dictionary)
        lda_model = gensim.models.ldamodel.LdaModel(corpus=corpus,
                                                    id2word=self.dictionary,
                                                    num_topics=self.numb_topics,
                                                    random_state=100,
                                                    update_every=1,
                                                    chunksize=100,
                                                    passes=10,
                                                    alpha='auto',
                                                    per_word_topics=True)

        return lda_model



 # /Diffusion/tlt.py
# Implementation of TLT model
import pathlib
import pickle
import numpy as np
from tqdm import tqdm, tqdm_notebook

from womg.diffusion.diffusion_model import DiffusionModel
from womg.utils.saver import TxtSaver


class TLT(DiffusionModel):
    '''
    Implementing Topic-aware Linear Threshold model

    Attributes
    ----------
    _numb_steps : int
    _numb_nodes : int
    _numb_docs : int
    _numb_topics : int
    new_active_nodes : dict
    active_nodes : dict
    inactive_nodes : dict

    Methods
    -------
    diffusion_setup
        sets all the attributes of the superclass and current class
    iteration
        iterates diffusion for a single step and save results
    parameter
        valuates the tlt parameter of the diffusion process
    update_sets
        updates the active and inactive sets of nodes and current activating nodes
    setnetwork_model
        set network_model attribute to the loaded pickle object of
    tltnetwork_model class
        settopic_model
    set topic_model attribute to the loaded pickle object of
        tlttopic_model class
    set_sets
        sets the initial configuration of the active and inactive sets of nodes
    stop_criterior
        defines a way to stop the run method of the superclass
    read_class_pickle
        reads a pickle file containing a class

    Notes
    -----
    -
    '''

    def __init__(self, network_model, topic_model,
                 path_out,
                 progress_bar, single_activator,
                 virality_resistance):
        super().__init__()
        self.network_model = network_model
        self.topic_model = topic_model
        self.saver = TxtSaver(path_out)
        self.new_active_nodes = {}
        self.active_nodes = {}
        self.inactive_nodes = {}
        self._propagations = []
        self._numb_nodes = 0
        self._numb_docs = topic_model.numb_docs
        self._numb_topics = topic_model.numb_topics
        self._single_activator = single_activator
        #print("single_activator:", single_activator)
        if progress_bar:
            self._progress_bar = tqdm_notebook
        else:
            self._progress_bar = tqdm
        self._thresholds_values = []
        self._virality_resistance = virality_resistance
        self.all_propagations = [[] for _ in range(self._numb_docs)]


    def save_threshold_values(self, path_out):
        with open(path_out+'/threshold_values.pickle', 'wb') as f:
            pickle.dump(self._thresholds_values, f)

    def diffusion_setup(self):
        '''
        Sets all the attributes of the current and super class
        '''
        self._numb_nodes = int(self.network_model.info['numb_nodes'])
        self._numb_docs = int(self.topic_model.numb_docs)
        self._numb_topics = int(self.topic_model.numb_topics)
        self._stall_count = {}
        for i in range(self._numb_docs):
            self._stall_count[i] = 0
        self.set_sets()


    def iteration(self, step):
        '''
        Iterates diffusion for a single step:
        1. For each item it looks for inactive nodes and calculates the tlt
        parameter for activation.
        2. Then it updates the sets of nodes
        3. Save results contained in public attributes: new activated nodes
        '''
        #print('step:',step)
        self.saver.save_propagation(self.propagations, step)
        for prop in self.propagations:
            item, node = (prop.replace('\n', '').split(' '))
            self.all_propagations[int(item)].append((step, int(node)))

        if step == 0:
            for item in range(self._numb_docs):
                new_active_nodes = self.active_nodes[item]
                self.update_sets(item, new_active_nodes)
        for item in self._progress_bar(range(self._numb_docs)):
            new_active_nodes = set()
            if self.active_nodes[item] != set() or None:
                for node in self.inactive_nodes[item]:
                    if self.parameter(item=item, node=node):
                        new_active_nodes.add(node)
                self.update_sets(item, new_active_nodes)



    def parameter(self, item, node):
        '''
        Calculate the tlt diffusion model parameter for a node on a item.
        [ Topic-aware Linear threshold model ]
             W(item, node) = topic_distr * ∑_v weight_v,node
        parameter is equal to the scalar product of topic distribution of the
        item with the sum over active nodes with out link directed to the
        given node.
        If this parameter is greater or equal to the threshold then parameter()
        returns True which means the given node will activate.
            threshold = 1/item_virality
        Parameters
        ----------
        item : int
            index of the item
        node : int
            index of the node
        Returns
        -------
        bool
            True if parameter is greater or equal to threshold, which means node
            is activating on that given item. Else False.
        '''
        #print(self.topic_model.viralities[item])
        # threshold = 1/ #+ self._threshold_bias)
        # calculating the sum over active nodes on item linked with user

        v_sum = np.zeros(self._numb_topics)
        node_check = False

        if self.network_model._directed:
            neighbors = [u for u, _v in list(self.network_model._nx_obj.in_edges(node))]
        else:
            neighbors = [v for _u, v in list(self.network_model._nx_obj.edges(node))]

        for v in neighbors:
            if v != node and  v in self.active_nodes[item]:
                v_sum += np.array(self.network_model.graph[(v, node)])
                if not all(np.isfinite(v_sum)):
                    print('ATTENTION v_sum: ', v_sum, ' weight: ', self.network_model.graph[(v, node)], ' v: ', v, ' node: ', node)
                node_check = True
        if node_check:
            z_sum = np.dot(self.topic_model.items_descript[item], v_sum)
            self._thresholds_values.append((z_sum, self.topic_model.viralities[item]))
            if not np.isfinite(z_sum):
                print('ATTENTION node: ', node, 'z_sum: ', z_sum)
            return z_sum > self._virality_resistance * self.topic_model.viralities[item]
        else:
            return False



    def update_sets(self, item, new_active_nodes):
        '''
        Updating the active, inactive and current new active sets attributes.

            The value correspondent to 'item' key of the new_active_nodes
            dictionary is set equal to the given input set (new_active_nodes).
            This set is given by the iteration method where update_sets is called.

            The inactive_nodes attribute is updating discarding the activated
            nodes calculated inside iteration()

            The active_nodes attribute is updating adding the activated
            nodes calculated inside iteration()

        Parameters
        ----------
        item : int
            index of the item
        new_active_nodes : list of int
            list of the nodes that has just activated by tlt parameter evaluation
            in parameter() method called by iteration()

        Notes
        -----
        this method is called inside the items for-loop inside iteration() method

        '''
        self.new_active_nodes[item] = new_active_nodes

        if self.new_active_nodes[item] == set():
            #print('item ', item)
            self._stall_count[item] += 1

        removing_list = new_active_nodes.union(self.active_nodes[item])  ### needs improvement
        if removing_list != set():
            for node in removing_list:
                self.inactive_nodes[item].discard(node)
                self.active_nodes[item].add(node)


    def set_sets(self):
        '''
        Defines the initial configuration of the sets.
            active nodes are defined with a random extraction defined in
            distributions module in utilities package

            inactive_nodes are all the nodes set

            current new activated nodes set is equal to the active one
        '''
        for item in range(self._numb_docs):
            self.active_nodes[item] = self.godNode_influence_config(item)
            if self.active_nodes[item] == set():
                self._stall_count[item] += 1
            self.inactive_nodes[item] = set(self.network_model._nx_obj.nodes())

            self.new_active_nodes[item] = self.active_nodes[item]


    def stop_criterior(self):
        '''
        Stops the run if there are not new active nodes for given time step seq
        '''
        stall_factor = True
        for item in self._stall_count.keys():
            stall_factor *= (self._stall_count[item] >= 1)
        return stall_factor


    def godNode_influence_config(self, item):
        '''
        Returns the activated nodes for the initial configuration for a given item;
        the god node (connected with all nodes) influences all the others for the
        given item evaluating the tlt parameter
        Parameters
        ----------
        item : int
            item index
        Returns
        -------
        list of active nodes for the given item
        '''
        actives_config = []
        max_interested = -np.inf
        max_v = None
        for u, v in self.network_model.godNode_links:
            curr_weight = self.network_model.graph[(u, v)]
            z_sum = np.dot(self.topic_model.items_descript[item], curr_weight)
            self._thresholds_values.append((z_sum, self.topic_model.viralities[item]))
            #print('z_sum: ', z_sum, '  virality: ', self._virality_resistance *self.topic_model.viralities[item])
            if z_sum > self._virality_resistance * self.topic_model.viralities[item]:
                if self._single_activator:
                    if max_interested < z_sum:
                        max_interested = z_sum
                        max_v = v
                else:
                    if all(v == 0 for v in curr_weight):
                        print('ATTENTION nodes: ', u, v, '  curr_weight:', curr_weight, ' z_sum:', z_sum)
                    actives_config.append(v)

        if self._single_activator:
            if max_v is not None:
                actives_config.append(max_v)

        return set(actives_config)


    @staticmethod
    def list_format(dictio):
        '''
        Static method for converting the dict format of the propagations
        into a list format:
            'doc; activating_node \n'
        '''
        lista = []
        for key in dictio.keys():
            for node in dictio[key]:
                lista.append(str(key) + ' ' + str(node) + '\n')
        return lista

    @property
    def propagations(self):
        '''
        When propagations is called for save it is set to the current config
        '''
        self.propagations = self.list_format(self.new_active_nodes)

        return self._propagations

    @propagations.setter
    def propagations(self, value):
        self._propagations = value


    def read_class_pickle(self, model):
        '''
        Read the pickle file containing a class model instance
        Parameters
        ----------
        model : string
            'topic' or 'network'
        Returns
        -------
        Loaded object of the correspondent class: TLTNetworkModel or TLTTopicModel
        '''
        file = pathlib.Path.cwd() / str('__'+model+'_model')
        with open(file, 'rb') as f:
            rfile = pickle.load(f)
        return rfile



import numpy as np

class Propagation:
    def __init__(self, network_model, topic_model, diffusion_model):
        self.diffusion_model = diffusion_model
        self.topic_model = topic_model
        self.network_model = network_model

        self._propagations = None
        self._docs = None
        self._topic_distributions = None
        self._interests = None


    @property
    def propagations(self):
        '''
        una lista di M tuple (tempo, nodo)
        '''
        self.propagations = self.diffusion_model.all_propagations
        return self._propagations

    @propagations.setter
    def propagations(self, value):
        self._propagations = value

    @property
    def docs(self):
        self.docs = [v for k, v in self.topic_model.items_keyw.items()]
        return self._docs

    @docs.setter
    def docs(self, value):
        self._docs = value


    @property
    def topic_distributions(self):
        '''
        KxM matrix
        '''
        vectors = [v for v in self.topic_model.items_descript.values()]
        self.topic_distributions = np.column_stack(vectors)
        return self._topic_distributions

    @topic_distributions.setter
    def topic_distributions(self, value):
        self._topic_distributions = value

    @property
    def interests(self):
        '''
        kxN matrix
        '''
        interests = [v for v in self.network_model.users_interests.values()]
        self.interests = np.column_stack(interests)
        return self._interests

    @interests.setter
    def interests(self, value):
        self._interests = value



# /network/tn.py
# Implementation of the Topic-aware Network model
import os
import womg
import random
import pathlib
import collections
import networkx as nx
import numpy as np
from scipy import sparse
from tqdm import tqdm, tqdm_notebook
from scipy.spatial import distance
from sklearn.decomposition import NMF
from womg.n2i.__main__ import n2i_nx_graph
from womg.network.tlt_network_model import TLTNetworkModel
from womg.utils.utility_functions import read_edgelist
from womg.utils.distributions import random_powerlaw_vec



class TN(TLTNetworkModel):
    '''
    Implementation of interest-influence model

    Attributes
    ----------
    - users_interests : dict
        dictionary containing the  interests vectors
        (of _numb_topics dimension) in the format:
        key <- [node id]
        value <- _numb_topics dimension array in numpy format
    - users_influence : dict
        dictionary containing the  influence vectors
        (of _numb_topics dimension) in the format:
        key <- [node id]
        value <- _numb_topics dimension array in numpy format
    - _nx_obj : NetworkX object
        networkx instance of the input network
    - godNode_links : dict
        dictionary containing all the links of the god node;
        god node is out connected
        to all the nodes but does not have in connections;
        god node index (id) is -1; format will be:
        key <- (-1, node id) [all int]
        value <- link weight [int]
    - _numb_topics : int
        dimension of the interests and influence vectors
    - _fast : bool
        flag for defining the chosen method for interests generation.
        if True the fastest (random) method is chosen

    Methods
    -------
    - set_graph()
    - set_godNode_links()
    - set_interests()
    - set_influence()
    - node2interests(): generates realistic latent interests of the nodes starting
      from the fixed given network
    - graph_weights_vecs_generation(): updates network weights using interests
      and influence vecs

    Notes
    -----
    This class implementation inherits from the abstract class networkModel()

    References
    ----------

    '''

    def __init__(self, numb_topics, homophily,
                 weighted, directed,
                 graph,
                 interests_path,
                 gn_strength,
                 infl_strength,
                 p, q, num_walks,
                 walk_length,
                 window_size, workers, iiter,
                 beta, norm_prior,
                 alpha_value, beta_value,
                 prop_steps=5000,
                 progress_bar=False,
                 seed=None
                 ):
        super().__init__()
        self._graph = graph
        self._weighted = weighted
        self._directed = directed
        self.users_interests = {}
        self.users_influence = {}
        self._numb_topics = numb_topics
        self._homophily = homophily
        self._interests_path = interests_path
        self.godNode_links = {}
        assert gn_strength is None or np.isfinite(gn_strength)
        self._godNode_strength = gn_strength
        self._infl_strength = infl_strength
        self._rand = 16 -15.875*self._homophily
        self._p = p
        self._q = q
        self._num_walks = num_walks
        self._walk_length = walk_length
        self._window_size = window_size
        self._workers = workers
        self._iiter = iiter
        self._beta = beta
        self._norm_prior = norm_prior
        self._alpha_value = alpha_value
        self._beta_value = beta_value
        if progress_bar:
            self._progress_bar = tqdm_notebook
        else:
            self._progress_bar = tqdm
        self._prop_steps = prop_steps
        self._seed = seed



    def network_setup(self, int_mode):
        '''
        - Sets the graph atribute using set_graph() method
        - Sets the info attribute using set_info() method
        - Sets the godNode_links attribute using set_godNode_links() method
        - Sets the interests vectors using set_interests() method
        - Sets the influence vectors using set_influence() mehtod
        - Sets the new graph weights using update_weights() method

        Notes
        -----
        See each method docstring for details
        '''
        if isinstance(self._graph, nx.classes.graph.Graph):
            self._nx_obj = self._graph
            self._directed = self._nx_obj.is_directed()
            self._weighted = nx.is_weighted(self._nx_obj)
            self.mapping = None
        elif self._graph == None:
            print('No graph path provided \n DEMO Mode: generating cascades in les miserables network')
            self._graph = pathlib.Path(os.path.abspath(womg.__file__).replace('/womg/__init__.py', '')) / "womgdata" / "graph" / "lesmiserables" / "lesmiserables_edgelist.txt"
            self._nx_obj, self.mapping  = read_edgelist(self,path=self._graph, weighted=False, directed=False)
        else:
            self._graph = pathlib.Path(self._graph)
            self._nx_obj, self.mapping = read_edgelist(self, path=self._graph, weighted=self._weighted, directed=self._directed)
        self.set_graph()
        if self._godNode_strength is not None:
            self.set_godNode_links()
        self.set_interests(int_mode)
        ##### check inf ##########################################
        for node, interests in self.users_interests.items():
            if not all(np.isfinite(interests)):
                print('ATTENTION  node: ', node, '  interests: ', interests)
        ###########################################################
        self.set_influence()
        #print('updating weights')
        self.graph_weights_vecs_generation()
        #print('Macroscopic homophily level: ', self.homophily(), ' with H=', self._homophily)


    def set_graph(self):
        '''
        Sets the graph attribute formatting the networkx instance with gformat()
        method of the superclass
        '''
        if isinstance(self._nx_obj, nx.classes.graph.Graph):
            self.graph = self.gformat(self._nx_obj, directed=self._directed)
            self.set_info()
        else:
            print('Not a networkx readable object')


    def set_info(self):
        '''
        Sets graph info dictionary attribute using networkx graph-instance description
        '''
        infos = nx.info(self._nx_obj) + '\nDirected: '+str(nx.is_directed(self._nx_obj))
        infos = infos.split()
        self.info['type'] = infos[2]
        self.info['numb_nodes'] = infos[6]
        self._numb_nodes = int(infos[6])
        self.info['numb_edges'] = infos[10]
        if infos[2] == 'MultiDiGraph':
            self.info['aver_in_degree'] = infos[14]
            self.info['aver_out_degree'] = infos[18]
            self.info['directed'] = infos[20]
        else:
            self.info['aver_degree'] = infos[13]
            self.info['directed'] = infos[15]



    def set_godNode_links(self, weight=1, nodes=None):
        '''
        Sets the godNode's links weights in the graph

        Parameters
        ----------
        weight : int
            scalar weight for the link/links
            (Default 1)

        nodes : iterable/int
            nodes indexes or node index on which one wants to change link weights
            if None is given all link weights (from godNode to each node) is changed
            (Defalut None)

        Example:
        tn_instance.set_godNode_links() :
        all the links weights (from godNode to each node) are set to 1
        '''
        if nodes is None:
            #print('Setting god node')
            for node in (self._nx_obj.nodes()):
                self.godNode_links[(-1, node)] = np.abs(np.random.randn())
                self.graph.update(self.godNode_links)
        if isinstance(nodes, int):
            self.godNode_links[(-1, nodes)] = weight
            self.graph.update(self.godNode_links)
        if isinstance(nodes, collections.Iterable):
            for node in nodes:
                self.godNode_links[(-1, node)] = weight
                self.graph.update(self.godNode_links)

    def set_interests(self, int_mode):
        '''
        Creates interests vectors (numb_topics dimension) for each node and
        save them in the users_interests class attribute.

        Parameters
        ----------
        - method : string
          name of the method for creating interests vectors
        '''
        print('Creating interests..')
        if int_mode == 'rand':
            #print('Random generation of interests:')
            self.random_interests()
        if int_mode == 'n2i':
            #print('\nGenerating interests from graph in ')
            self._q = np.exp(4.60517*self._homophily)
            self._p = np.exp(-4.60517*self._homophily)
            prior = 'half_norm' if self._norm_prior else 'beta'
            emb = n2i_nx_graph(
                        nx_graph=self._nx_obj,
                        window_size=self._window_size,
                        walk_length=self._walk_length,
                        num_walks=self._num_walks,
                        dimensions=self._numb_topics,
                        p=self._p,
                        q=self._q,
                        beta=self._beta,
                        prior=prior,
                        alpha_value=self._alpha_value,
                        beta_value=self._beta_value,
                        seed=self._seed
                   )
            for node in self._nx_obj.nodes():
                self.users_interests[node] = emb[node]

        if int_mode == 'prop_int':
            self.propag_interests()

        if int_mode == 'nmf':
            self.nmf_interests()

        if int_mode == 'load' or self._interests_path != None:
            print('Loading interests from: ', self._interests_path)
            self.load_interests()


    def set_influence(self):
        '''
        Sets influence vectors for all nodes

        Parameters
        ----------
        - method : string
          name of the method for creating interests vectors
        '''

        if self._infl_strength is None:
            for node in self._nx_obj.nodes():
                self.users_influence[node] = [0. for _ in range(self._numb_topics)]
        else:
            fitness = 1 + np.random.pareto(a=self._infl_strength, size=self._nx_obj.number_of_nodes())
            if np.any(np.isinf(fitness)):
                print('Wrong parameter infl strenght')
            fitness = np.minimum(fitness, 10E20)
            for node in self._nx_obj.nodes():
                self.users_influence[node] = fitness[node] * self.users_interests[node]

        if self._godNode_strength is not None:
            self.users_influence[-1] = np.ones(self._numb_topics)

        for node, value in self.users_influence.items():
            if not all(np.isfinite(value)):
                print('ATTENTION node: ', node, ' influences: ', value)
        '''
        random_powerlaw_vec(gamma=self._rho, self._numb_topics)
        # rescaling infleunce importance
        norm_avg = 0.
        for node in self._nx_obj.nodes():
            norm_avg += np.linalg.norm(self.users_interests[node])/self._numb_nodes
        scale_fact = self._infl_strength*norm_avg
        # setting influence vec
        for node in self._nx_obj.nodes():
            influence_vec = self.node2influence(scale_fact)
            self.users_influence[node] = influence_vec
        '''

    def random_interests(self, norm=True):
        '''
        Create interests vector for each node using a random uniform probability density
        function and directly saves interests vectors in attribute users_interests

        Parameters
        ----------
        - norm : bool
        if True interests vectors are normalized

        '''
        if norm:
            for node in self._nx_obj.nodes():
                self.users_interests[node] = np.random.rand(self._numb_topics)


    def propag_interests(self):
        '''
        generates interests with the propagation method starting from "n" nodes
        '''
        # INITIALIZE
        for i in list(self._nx_obj.nodes):
            interests = np.random.dirichlet(np.ones(self._numb_topics)*1./self._numb_topics)
            self.users_interests[i] = interests

        # SELECT INFLUENCERS
        # start with a random node
        influencers = [np.random.choice(self._nx_obj.nodes)]

        for i in range(14):
            # calculate distances to current influencers
            sp = {i: nx.shortest_path(self._nx_obj, i) for i in influencers}
            distances = np.array([list(len(sp[j][i]) for j in influencers) for i in self._nx_obj.nodes()])
            # select the node fartest from all influencers
            influencers.append(distances.min(axis=1).argmax())

        # PROPAGATION STEP
        for c in range(self._prop_steps):
            i = np.random.choice(self._nx_obj.nodes())
            interests_i = self.users_interests[i]
            lr = 0.5 if i in influencers else 0.01
            #lr = 0.1
            for j in list(self._nx_obj.neighbors(i)):
                if j in influencers:
                    continue
                interests_j = self.users_interests[j]
                interests_j += interests_i * lr
                interests_j /= interests_j.sum()
                self.users_interests[j] = interests_j



        # def overlap_generator(self):
    #     """
    #     Function to generate a neighbourhood overlap matrix (second-order proximity matrix).
    #     :param G: Graph object.
    #     :return laps: Overlap matrix.
    #     """
    #     G = self._nx_obj
    #     degrees = nx.degree(G)
    #     sets = {node:set(G.neighbors(node)) for node in nx.nodes(G)}
    #     laps = np.array([[float(len(sets[node_1].intersection(sets[node_2])))/(float(degrees[node_1]*degrees[node_2])**0.5) if node_1 != node_2 else 0.0 for node_1 in nx.nodes(G)] for node_2 in nx.nodes(G)],dtype = np.float64)
    #     return laps

    def overlap_generator(self, A):
        """
        Generate the second order overlap from a sparse adjacency matrix A.
        """
        aat = A.dot(A.T)
        d = aat.diagonal()
        ndiag = sparse.diags(d, 0)
        n = np.sqrt(ndiag.dot(aat>0).dot(ndiag))
        n.data[:] = 1./n.data[:]
        return aat.multiply(n) #- sparse.identity(aat.shape[0])


    # def nmf_interests(self, eta=64.):
    #     '''
    #     Generates interests according to non-negative matrix factorization
    #     method
    #
    #     Parameters
    #     ----------
    #
    #     Returns
    #     -------
    #
    #     '''
    #     print('\n Generating interests with nmf ..')
    #     A = nx.to_numpy_matrix(self._nx_obj)
    #     S_0 = self.overlap_generator()
    #     R = np.random.rand(self._nx_obj.number_of_nodes(), self._nx_obj.number_of_nodes())
    #
    #     S = eta*S_0 + A + self._rand*R
    #     model = NMF(n_components=self._numb_topics, init='random', random_state=self._seed)
    #     W = model.fit_transform(S)
    #
    #     for node in self._nx_obj.nodes():
    #         self.users_interests[node] = W[node]

    def nmf_interests(self, eta=64.):
        #beta = self._homophily
        A = nx.adjacency_matrix(self._nx_obj)
        S_0 = self.overlap_generator(A)
        R = np.random.rand(self._nx_obj.number_of_nodes(), self._nx_obj.number_of_nodes())
        #S = beta*(S_0 + A + sparse.identity(A.shape[0])) + (1-beta)*R
        eta = 64.
        S = eta*S_0 + A + self._rand*R
        model = NMF(n_components=self._numb_topics, init='nndsvd', random_state=self._seed)
        W = model.fit_transform(S)
        #print('Doing nmf with random coeff ', self._rand, ' and homophily ', self._homophily)
        if not np.all(np.isfinite(W)):
            print('ATTENTION W contains infinites')

        for node in self._nx_obj.nodes():
            self.users_interests[node] = W[node]

    def node2influence(self, scale_fact, alpha_max=10):
        '''
        Creates influence vectors (numb_topics dimension) from dirichlet
        distribution over topics for a single node

        Parameters
        ----------
        - alpha_max : int
          highest value in the dirichlet weight vec for a random entry
        '''
        dirich_weight_vec = [1 for topic in range(self._numb_topics)]
        dirich_weight_vec[random.randint(0, self._numb_topics-1)] = alpha_max
        influence_vec = np.random.dirichlet(dirich_weight_vec)
        return scale_fact*influence_vec


    def graph_weights_vecs_generation(self):
        '''
        Creates weights vectors (numb_topcis dimension) for each direct link
        and update the graph attribute; a link weight is defined as:

        w_(i,j) = intersts_j + influence_i

        Parameters
        ----------
        - god_node_weight : int
          scalar value for each entry of the weight vector involving the godNode
        '''
        for link in self.graph.keys():
            # god node
            if link[0] == -1:
                out_influence_vec = self.users_influence[-1]
                in_interest_vec = self.users_interests[link[1]]
                self.set_link_weight(link, in_interest_vec + self._godNode_strength * out_influence_vec)
            else:
                out_influence_vec = self.users_influence[link[0]]
                in_interest_vec = self.users_interests[link[1]]
                self.set_link_weight(link, out_influence_vec + in_interest_vec)


    def set_link_weight(self, link, new_weight):
        '''
        Sets the link attribute to the given weights vector arg

        Parameters
        ----------
        - link : tuple
          link in the graph in which you want to change the attribute
        - new_weight : array
          numb_topic dimension array that is going to be the new attribute of the
          link
        '''
        if not all(np.isfinite(new_weight)):
            print('ATTENTION link ', link, ' weight: ', new_weight)
        self.graph[link] = new_weight


    def load_interests(self):
        '''
        Loads interests vector from path
        '''
        with open(self._interests_path, 'r') as f:
            for _ in f.readlines():
                node, interests = _.split(' ', 1)[0], _.split(' ', 1)[1]
                self.users_interests[int(node)] = np.array(eval(interests[:-1]))




    ################# Analysis

    def sim_in(self):
        sims = []
        for i in self._nx_obj.nodes:
            for j in list(self._nx_obj.neighbors(i)):
                sims.append(1 - distance.cosine(self.users_interests[i], self.users_interests[j]))
        return np.mean(sims)

    def select_notedge(self):
        v1 = np.random.choice(self._nx_obj.nodes())
        v2 = np.random.choice(self._nx_obj.nodes())

        while (v1,v2) in self._nx_obj.edges or v1==v2:
            v1 = np.random.choice(self._nx_obj.nodes())
            v2 = np.random.choice(self._nx_obj.nodes())
        return v1, v2

    def sim_out(self, samples):
        sims_out = []
        for c in range(samples):
            i, j = self.select_notedge()
            sims_out.append(1 - distance.cosine(self.users_interests[i], self.users_interests[j]))
        return np.mean(sims_out)

    def homophily(self, numb_not_edges_tested=10000):
        return self.sim_in() / self.sim_out(numb_not_edges_tested)
